/* NAME:
 *	domaincmp - compare domains
 *
 * SYNOPSIS:
 *	int domaincmp("domain", "spec", "flags")
 *
 * DESCRIPTION:
 *	This routine, tests whether "domain" matches "spec".
 *	If "spec" contains no special chars or only a leading ``*''
 *	then a simple strcasecmp(3) is used.
 *
 *	If "spec" is a litteral that starts with ``.'' then it will
 *	matched against the tail of "domain".
 *	
 *	Otherwise if "spec" is a literal,  it must match "domain"
 *	exactly or if bit 1 of "flags" is set ".spec" must be a tail
 *	match for "domain". 
 *	
 *	If "spec" more complicated, it is translated into a POSIX
 *	regular expression, and then tested against "domain".
 *
 * BUGS:
 *	Not everyone has the POSIX regcomp,regexec routines.
 *	Fortunately common usage does not require regexps.
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1997, Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */

#ifndef lint
static char RCSid[] = "$Id: domaincmp.c,v 1.6 1998/07/10 07:12:36 sjg Exp $";
#endif
#include <sys/cdefs.h>
#include <stdio.h>
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <sys/param.h>
#include <sys/types.h>
#include <string.h>
#if defined(BSD4_4) || defined(HAVE_REGEXEC)
# include <regex.h>			/* POSIX 1003.2 */
#endif

int
domaincmp(domain, spec, flags)
     const char *domain, *spec;
     int flags;
{
	char re[264];
	const char *cp = 0;
	int dl, sl, n, i, j, s;
#if defined(BSD4_4) || defined(HAVE_REGEXEC)
	regex_t reg;
#endif
	
	sl = strlen(spec);
	dl = strlen(domain);
	/*
	 * if spec contains any wild-cards etc, then we need to
	 * do some work, otherwise we just do a simple compare.
	 */
	if ((cp = strpbrk(spec, "*?[")) == 0)
		cp = spec;
	else if (*spec == '*' && strpbrk(&spec[1], "*?[") == 0) {
		cp = &spec[1];
		flags |= 1;			/* tail match ok  */
	}
	/*
	 * if "spec" is just domain, *domain or *.domain, it is easy,
	 * otherwise we need to compile a regexp (if we can).
	 */
	if (cp) {
		if ((i = strcspn(cp, ":")) > 0) {
			/*
			 * spec is domain:port, we ignore the port here
			 */
			strncpy(re, cp, i);
			re[i] = '\0';
			cp = re;
			sl -= (sl - i);
		}
		i = strlen(cp);
		if (dl < i)
			return -1;	/* can't match */
		else if (dl > i) {
			if (*cp != '.' && *spec != '*' && domain[dl - i - 1] != '.')
				return -2;	/* don't match foo as ufoo */
		}
		s = strcasecmp(&domain[dl - i], cp);
		if (s == 0 &&			/* matched */
		    *cp != '.' &&		/* .spec should always be ok */
		    dl > sl &&			/* tail match */
		    (flags & 1) == 0)		/* tail match not allowed */
			s = -3;
		return s;
	}
#if defined(BSD4_4) || defined(HAVE_REGEXEC) || defined(HAVE_RE_COMP)
	/*
	 * wow, something complex, time for a regexp?
	 */
	re[0] = '^';
	for (i = 0, j = 1; i < sl; i++) {
		switch (spec[i]) {
		case ':':			/* domain:port */
			sl = i;			/* that's the end */
			break;
		case '.':
			re[j++] = '\\';
			re[j++] = '.';
			break;
		case '?':
			re[j++] = '.';
			break;
		case '*':
			re[j++] = '.';
			re[j++] = '*';
			break;
		default:
			re[j++] = spec[i];
			break;
		}
	}
	re[j++] = '$';
	re[j] = '\0';

#if defined(BSD4_4) || defined(HAVE_REGEXEC)
	if (regcomp(&reg, re, REG_ICASE|REG_NOSUB|REG_EXTENDED) == 0) {
		s = regexec(&reg, domain, 0, 0, 0);
		return s;
	}
#elif defined(HAVE_RE_COMP)
	if (re_comp(re) == 0)
		return (re_exec(domain) != 1);
#endif
#endif
	return -4;
}

#ifdef MAIN
int
main(argc, argv)
     int argc;
     char **argv;
{
	int i;

	for (i = 2; i < argc; i++) {
		printf("domaincmp(%s,%s,0)->%d\n",
		       argv[i], argv[1], domaincmp(argv[i], argv[1], 0));
		printf("domaincmp(%s,%s,1)->%d\n",
		       argv[i], argv[1], domaincmp(argv[i], argv[1], 1));
	}
	exit(0);
}
#endif

	
