/* NAME:
 *	tcp_open - connect a socket to server
 *
 * SYNOPSIS:
 *	int	net_open(const char *spec, const char *mode)
 *	int	tcp_open(const char *host, int port)
 *
 * DESCRIPTION:
 *	This routine simply gathers and hides all the nasty details of setting
 *	up a socket and connecting to a server.  Note that we deliberately
 *	ignore the issue of SIGPIPE here.  Many appls should terminate if they
 *	get a SIGPIPE, others may want to do clever things and must make their
 *	own arrangements anyway.
 *
 *	"Spec" is fodder for sock_dest(3) such as ``tcp!host:port''.
 *
 *	"Mode" indicates whether we should simply 'b' bind(2) the port or
 *	'c' connect(2) to the destination identified by "spec".  If
 *	"mode" starts with 'r' we bindresvport(3) before the connect.
 *	If bindresvport(3) fails but "mode" starts with 'ro' we carry
 *	on regardless.  Finally if "mode" contains :"tries" then we
 *	will retry the connect(2) that many times with an increasing
 *	delay between retries.  Note that if the destination has
 *	multiple addresses, each will be tried before delaying.
 *
 *	tcp_open() simply calls net_open().
 *
 * RETURNS:
 *	socket or -1 on failure.
 *
 * AUTHOR:
 *	Simon Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */

#ifndef lint
static char RCSid[] = "$Id: tcpopen.c,v 1.7 1998/08/16 05:35:30 sjg Exp $";
#endif
#include <stdio.h>
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>

#include "tcp.h"

#ifdef SOCKS
# include "socks.h"
#endif

extern int errno;

/*
 * "spec" is like [{udp,tcp}!]host:port
 * "mode" is "b*" for bind or anything else means connect.
 *	"r[o]" for bindresvport [optional]
 *	"*:tries" overrides default of 5.
 */

int	
net_open(spec, mode)
	const char *spec;
	const char *mode;
{
	struct sockaddr_in	sa;
	char			**addr_list, **addr;
	unsigned long		ip_addr = 0;
	int			s = -1, sock_type = SOCK_STREAM;
	int timo, try, tries = 6;
	int last_error = 0;
	register char *rcp;

	if (mode == 0 || *mode == '\0')
		mode = "c";		/* connect */
	
	if ((addr_list = sock_dest(spec, &sa, &sock_type)) != 0) {
		if (*mode == 'b') {	/* bind */
			if ((s = socket(AF_INET, sock_type, 0)) < 0) {
				syslog(LOG_ERR, "net_open: socket(): %m");
				perror("socket");
				return -1;
			}
			if (mode[1] == 'r') {
				int b = 1;
				
				if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
					       (void *)&b, sizeof (int)) < 0) {
					syslog(LOG_WARNING, "net_open: setsockopt(%d, SO_REUSEADDR): %m", s);
					perror("SO_REUSEADDR");
				}
			}
			if (bind(s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
				syslog(LOG_ERR, "net_open: bind(%d, '%s'): %m",
				       s, ip_caller_id(&sa, 0));
				perror("bind");
				close(s);
				return -1;
			}
			return s;
		}
		/*
		 * connect, trying each address in turn.
		 */
		if ((rcp = strchr(mode, ':')) != 0)
			tries = atoi(++rcp);
		for (try = timo = 1; try <= tries; try++) {
			for (addr = addr_list; addr && *addr; addr++) {
				ip_addr = *((unsigned long *) *addr);

				if (ip_addr == 0 || ip_addr == (unsigned long) -1L)
					continue; /* try next */
				/*
				 *	We need a fresh socket for each
				 *	connect attempt.
				 *	If we fail to get a socket, there is
				 *	no point continuing.
				 */
				if ((s = socket(AF_INET, sock_type, 0)) < 0)
					return (-1);
				if (*mode == 'r' && /* reserved */
				    bindresvport(s, (struct sockaddr_in *) 0) < 0) {
					if (mode[1] != 'o') {
						/* not optional */
						close(s);
						return (-1);
					}
				}
				sa.sin_addr.s_addr = ip_addr;
				
				if (sock_type != SOCK_STREAM)
					return s;
				
				if (connect(s, (struct sockaddr *) &sa,
					    sizeof(struct sockaddr_in)) == 0)
					return s;
				last_error = errno;
				close(s);
				if (Debug) {
					syslog(LOG_DEBUG, "net_open: %s: %m",
					       spec);
					perror(spec);
				}
			}
			switch (last_error) {
			case ECONNREFUSED:
			case EHOSTUNREACH:
			case ENETDOWN:
			case ENETUNREACH:
				break;
			default:
				try = tries + 1;
				break;
			}
			if (try < tries) {
				sleep(timo);
				timo *= 2;
			}
		}
	}
	if (last_error)
		errno = last_error;	/* REVISIT */
	return (-1);
}

int	
tcp_open(host, port)
	const char *host;
	int port;
{
	char spec[256];

	sprintf(spec, "tcp!%.200s:%d", host, port);

	return net_open(spec, "connect");
}

#ifdef MAIN
int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_in sa;
	int x, s;

	Debug = 1;
	if ((s = net_open(argv[1], (argc > 2) ? argv[2] : "c")) >= 0) {
		x = sizeof(sa);
		if (getpeername(s, (struct sockaddr *) &sa, &x) == 0) {
			printf("connected to %s\n",
			       ip_caller_id(&sa, 2));
		}
		copy(s, 1);
		close(s);
	} else {
		perror("net_open");
	}
	exit (s < 0);
}
#endif
