/*
 * NAME:
 *	ssld - routines useful for SSL servers
 *
 * SYNOPSIS:
 *	int ssld_init(const char *"cert_file",
 *			const char *"key_file",
 *			const char *"cipher_list",
 *			int "verify_flag", int "flags")
 *	void set_ssldreply(void (*"f")())
 *      int ssld_auth_cert(const char *"key", const char *"cert",
 *			const char *remhost, const char *"auth_files")
 *	int ssl_cert_ok(const char *"key", const char *"cert",
 *			const char *remhost)
 *	int ssld_auth(int "s", const char *"remuser",
 *			const char *"locuser")
 *	int ssld_start(int "s", int "log_lev", int "active",
 *			int (*"cb")(), int "vf", void (*"reply")())
 *	int certcmp(const char *"cert", char *"re")
 * 
 * DESCRIPTION:
 *	These routines are intended for use by SSL based daemons.
 *	Routines which are shared between clients and servers are
 *	documented in sslfd(3).
 *
 *	A server (such as 'SSLrshd') typically needs to call:
 *
 *	ssld_init("cert_file", "key_file", "cipher_list",
 *		"verify_flag", "init_flags");
 *
 *	To change any of the defaults for these arguments and tell
 *	SSLeay which verify callback function to use.  If command
 *	line args have been handled by ssl_setopt(3), and the
 *	application is using ssld_start(), then the call to
 *	ssld_init() may not be needed.  
 *
 *	To establish the SSL protocol, the server must call one of:
 *.nf
 *
 *	ssl_start("sock", "log_lev", "active")
 *	ssl_start_cb("sock", "log_lev", "active", "verify_cb", "verify_flag")
 *	ssld_start("sock", "log_lev", "active", "verify_cb", "verify_flag",
 *		"reply")
 *.fi
 *
 *	with "active" = 1 or 0.	The end with "active" = 0 will call
 *	SSL_accept() and must always provide a certificate.  For
 *	applications where client to server authentication is most
 *	important, it makes sense to reverse the accept/connect roles
 *	at the SSL level. 
 *
 *	ssld_start() is different to the others in that if
 *	"verify_flag" inlcudes the SSL_VERIFY_FAIL_IF_NO_PEER_CERT
 *	bit, it will not be passed to SSL.  Instead, if there is an
 *	error during setup (such as client cert has expired or been
 *	revoked), "reply"(error_msg) is called to inform the client of
 *	the reason for failure, before the SSL session is terminated.
 *	If "verify_cb" is NULL, the default which checks the issuer's
 *	CRL for revoked certificates is used.
 *	Most servers (except ftpd) should probably call ssld_start(),
 *	and almost never use a verify callback other than the default
 *	one in this module. 
 *
 *	Although originally written for 'SSLrshd' which is the server
 *	for ssl_rcmd(3), servers can call:
 *
 *	ssld_auth("socket", "remuser", "key")
 *
 *	to determine if the client certificate provides authorization
 *	to access some resource identified by "key".  In the case of
 *	'SSLrshd' "key" is the requested local user id.
 *
 *	ssld_auth() determines the remote "hostname".  If the connection
 *	came from the loopback address, then gethostname(2) is used
 *	otherwise getpeername(2).  Normally, "hostname" is only used
 *	for logging purposes and we do not care where the connection
 *	comes from.  However, if the certificate is a host cert (/CN
 *	is a domain name that we can find in the DNS), then we most
 *	certainly do care where the connection came from.  Also, we
 *	now support a host list appended to an entry in the "auth"
 *	files, such that a matching cert will only be accepted from
 *	the listed hosts.  For these reasons we always verify the
 *	hostname via the DNS and check that the calling address is
 *	listed.  If not, we use the remote hostname ``unknown''.
 *	
 *	ssld_auth() calls get_x509_name(0, 0) to extract the
 *	common name (/CN) field from the client's certificate.
 *	If the name contains any ``.'' it is compared against the
 *	remote hostname using domaincmp(3) and if that fails, we
 *	confirm that the cert name is either a wildcard host cert or
 *	can be found in the DNS via gethostbyname(2).  If the cert is
 *	indeed a host cert and domaincmp(3) failed, we reject the
 *	cert. 
 *
 *	If not already rejected, ssld_auth() calls:
 *
 *	ssl_cert_ok("key", "ssl_auth_cert", "hostname")
 *
 *	To lookup "ssl_auth_cert" in the  "auth" file(s).  If
 *	certcmp() indicates a match, and either "key" or 'root' is
 *	listed as a valid target, permission is granted,
 *	otherwise it is denied.  If the target 'deny' is found in the
 *	same record the certificate is denied access.
 *
 *	The format of the "auth" files is:
 *.nf
 *
 *	"user_list":[-e ]"cert"[:"host_list"]
 *.fi
 *		
 *	Which is backwards compatible with Tim Hudson's ssl.users
 *	file which is included in the default list of auth
 *	files (/etc/ssl.deny:/etc/ssl.root:/etc/ssl.local:
 *	/etc/ssl.users:/etc/ssl.global) and the search logic is very
 *	similar.   If "host_list" is supplied the matching "cert" will
 *	only be accepted from the listed hosts or domains.  This is
 *	useful when "cert" contains a wildcard, or the '-e' option is
 *	present in which case "cert" is treated as a regular
 *	expression by certcmp().  If the POSIX regexec(3) is available
 *	it is used and "cert" may be an extended regular (case
 *	insensitive), otherwise if re_comp(3) is available "cert" must
 *	be a basic regexp.  If "user_list" contains entries like /CN=,
 *	then the appropriate field is extracted from the cert.  This
 *	is only useful if "cert" was a regexp. Note that a field value
 *	of ``root'' will be skipped to ensure privileged access is not
 *	granted accidentally.  Certs to be granted ``root'' access
 *	must be explicitly listed.
 *
 * SEE ALSO:
 *	sslc(3), ssl_rcmd(3), ssl(3), netio(3).
 *
 * NOTES:
 *	The current version of SSLeay can be obtained from:
 *	'ftp://ftp.psy.uq.oz.au/pub/Crypto/SSL/SSLeay-x.xx.tar.gz'
 *
 *	The SSLeay FAQ by Tim Hudson <tjh@mincom.oz.au> can be found
 *	at 'http://www.psy.uq.oz.au/~ftp/Crypto'
 *
 * BUGS:
 *	The ftp protocol is such that ssld_start() cannot be used.
 *	The ftp client either does not display the reason why the
 *	session is being dropped or worse, just gets out of sync and
 *	hangs.   You can only take generic facilities so far...
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: ssld.c,v 1.42 1998/11/13 02:44:05 sjg Exp $";
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <syslog.h>
#include <string.h>
#if defined(BSD4_4) || defined(HAVE_REGEXEC)
# include <regex.h>			/* POSIX 1003.2 */
#endif

#include "x509.h"
#include "ssl.h"

#include "sslfd.h"

#ifdef SSLEAY8
# include "x509_vfy.h"
#else
/*
 * Since the 0.8 errs are a superset of the 0.6 ones, we use the new ones.
 */
# define X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT VERIFY_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
# define X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT VERIFY_ERR_UNABLE_TO_GET_ISSUER 
# define X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY VERIFY_ERR_UNABLE_TO_GET_ISSUER 
# define X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD VERIFY_ERR_ERROR_IN_NOT_BEFORE_FIELD 
# define X509_V_ERR_CERT_SIGNATURE_FAILURE VERIFY_ERR_SIGNATURE_FAILURE
# define X509_V_ERR_CERT_NOT_YET_VALID VERIFY_ERR_CERT_NOT_YET_VALID
# define X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD VERIFY_ERR_ERROR_IN_NOT_AFTER_FIELD
# define X509_V_ERR_CERT_HAS_EXPIRED VERIFY_ERR_CERT_HAS_EXPIRED
#endif

#ifndef MAXHOSTNAMELEN
# define MAXHOSTNAMELEN 64
#endif

extern char x509_name[256];

/*
 * This function derrived from some logic in Tim Hudson's
 * verify_callback()
 * The logic here is a little different - we only accept the
 * connection if the cert and user name match in our database.
 */
#ifndef _PATH_SSL_USERS
# define _PATH_SSL_USERS "/etc/ssl.deny:/etc/ssl.root:/etc/ssl.local:/etc/ssl.users:/etc/ssl.global"
#endif

/*
 * Compare "cert" against "re"
 * If "re" begins with '-e' we treat it as a real regexp
 * otherwise we just use strcasecmp().
 */

int
certcmp(cert, re)
	const char *cert;
	char *re;
{
	/*
	 * A bit of ugliness to handle the two decodings for
	 * oid OBJ_X509,8L used by SSLeay versions before and after
	 * 0.8.0
	 * We replace any "/ST=" in the auth files with "/SP=", unless
	 * we are compiled against SSLEAY8 in which case do the opposite.
	 */
#ifdef SSLEAY8
	char *s1 = "/SP=", *s2 = "/ST=";
#else
	char *s1 = "/ST=", *s2 = "/SP=";
#endif
	char *cp;
	int ret = -2;
	
	if (cert == 0 || re == 0)
		return -1;
	
	if ((cp = strstr(re, s1)) != 0) {
		(void)strncpy(cp, s2, 3);
	}
#if defined(DEBUG) || defined(MAIN)
	syslog(LOG_DEBUG, "certcmp(%s, %s)",
	       cert, re);
#endif
	if (re[0] == '-' && re[1] == 'e') {
#if defined(BSD4_4) || defined(HAVE_REGEXEC)
		regex_t reg;
#endif

		re += 2;
		while (*re == '\t' || *re == ' ')
			re++;

#if defined(BSD4_4) || defined(HAVE_REGEXEC)	
		if (regcomp(&reg, re, REG_ICASE|REG_NOSUB|REG_EXTENDED) == 0) {
			ret = regexec(&reg, cert, 0, 0, 0);
		} else
			syslog(LOG_ERR, "regcomp(%s) failed", re);
#elif defined(HAVE_RE_COMP)
		if (re_comp(re) == 0) {
			ret = (re_exec(cert) != 1);
		} else
			syslog(LOG_ERR, "re_comp(%s) failed", re);
#else
		syslog(LOG_ERR, "no REGEXPs, cannot match re='%s'", re);
#endif
		return ret;
	}
	return strcasecmp(cert, re);
}



int
ssld_auth_cert(key, cert, remhost, auth_files)
	const char *key;
	const char *cert;
	const char *remhost;
	const char *auth_files;
{
	char buf[1024];
	FILE *fh;
	int line = -1;
	char *cp, *n, *s, *h;
	char *file, *next_file, *files;

	int ret = SSLD_AUTH_NA;
	int match_result;			/* what does a match mean */
	
	if (!cert)
		return SSLD_AUTH_NA;

	if (!_path_ssl_users)
		_path_ssl_users = _PATH_SSL_USERS;
	if (!auth_files)
		auth_files = _path_ssl_users;

	files = next_file = strdup(auth_files);
	for (file = files; ret == 0 && file; file = next_file) {
		if (next_file = strchr(next_file, ':'))
			*next_file++ = '\0';
		
		if (fh = fopen(file, "r")) {
			line = 0;
			/*
			 * if file is foo.deny, any match means deny
			 */
			match_result = (strstr(file, ".deny") == 0) ? SSLD_AUTH_OK : SSLD_AUTH_DENY;
			while (ret == SSLD_AUTH_NA &&
			       fgets(buf, sizeof(buf), fh)) {
				line++;
				switch (buf[0]) {
				case '#':
				case '\n':
					continue;
				}
				if (cp = strchr(buf, '\n'))
					*cp = '\0';
				else {
					syslog(LOG_ERR, "%s: line %d, too long",
					       file, line);
					continue;
				}
				cp = strchr(buf, ':');
				if (!cp) {
					syslog(LOG_ERR, "%s: line %d, invalid format",
					       file, line);
					continue;
				}
				*cp++ = '\0';
				if ((h = strchr(cp, ':')) != 0)
					*h++ = '\0';
				if (certcmp(cert, cp) == 0) {
					if (match_result == SSLD_AUTH_DENY) {
						ret = SSLD_AUTH_DENY;
						continue;
					}
					if (h && *h && remhost && *remhost) {
						/*
						 * might be a host cert
						 * with wild-cards,
						 * in any case we will
						 * only accept it from
						 * the listed hosts.
						 */
						do {
							if (s = strchr(h, ','))
								*s++ = '\0';
							if (domaincmp(remhost,
								       h, 0) == 0) {
								remhost = 0;
								break;
							}
							h = s;
						} while (h) ;
						if (remhost) {
							/*
							 * it didn't match
							 */
							ret = SSLD_AUTH_WRONG_HOST;
							break;
						}
					}
					n = buf;
					ret = SSLD_AUTH_NA;
					while (n) {
						cp = strchr(n, ',');
						if (cp)
							*cp++ = '\0';

						if (*n == '/' &&
						    (s = get_x509_field(0, 0, n)) != 0 &&
						    strcasecmp(s, "root") != 0) {
							if ((h = strpbrk(s, "@")) != 0)
								*h = '\0';
						} else
							s = n;
						/*
						 * 'deny' means just that
						 */
						if (strcmp("deny", s) == 0) {
							ret = SSLD_AUTH_DENY;
						}
						/*
						 * if the cert is
						 * allowed root access
						 * then it is allowed any...
						 */
#if defined(DEBUG) || defined(MAIN)
						syslog(LOG_DEBUG,
						       "check: key(%s) == n(%s)",
						       key, n);
#endif
						if (key &&
						    strcasecmp(key, s) == 0 ||
						    strcasecmp("root", s) == 0) {
							if (ret == SSLD_AUTH_NA)
								ret = SSLD_AUTH_OK;
							break;
						}
						n = cp;
					}
				}
			}
			fclose(fh);
		}
	}
	free(files);

	if (line >= 0 && ret == SSLD_AUTH_NA) {
		/*
		 * we found at least one file in auth_files,
		 * but no matching entry for this cert.
		 */
		ret = SSLD_AUTH_REJECT;
	}
	switch (ret) {
	case SSLD_AUTH_NA:
		syslog(LOG_DEBUG, "SSL: no auth file(s): %s", auth_files);
		break;
	case SSLD_AUTH_WRONG_HOST:
		syslog(LOG_WARNING, "SSL: target '%s' cert '%s' from '%s' DENIED",
		       key, cert, remhost);
		break;
	case SSLD_AUTH_REJECT:
		syslog(LOG_WARNING, "SSL: target '%s' cert '%s' REJECTED",
		       key, cert);
		break;
	case SSLD_AUTH_DENY:
		syslog(LOG_ERR, "SSL: target '%s' cert '%s' DENIED",
		       key, cert);
		break;
	case SSLD_AUTH_OK:
		syslog(LOG_INFO, "SSL: target '%s' cert '%s' OK",
		       key, cert);
		break;
	}
	return ret;
}

int
ssl_cert_ok(key, cert, remhost)
	const char *key;
	const char *cert;
	const char *remhost;
{
	return ssld_auth_cert(key, cert, remhost, 0);
}

static int cert_revoked = 0;

static void
_rply(msg)
	char *msg;
{
	write_net(ssl_fd, msg, strlen(msg));
}

void (*ssld_reply)() = _rply;

void
set_ssldreply(f)
	void (*f)();
{
	if (!f)
		f = _rply;
	ssld_reply = f;
}

int
ssld_auth(s, remuser, key)
	int s;
	const char *remuser, *key;
{
	struct sockaddr_in from;
	struct hostent *hp;
	int i;
	char *hostname;
	char hostaddr[MAXHOSTNAMELEN + 1];
	char hostbuf[2 * MAXHOSTNAMELEN + 1];
	char msg[512];
	int sts = SSLD_AUTH_REJECT;
	char *cert_name;
	char *cp, *caller = 0;
	char *faked = 0;
	
	msg[0] = 0;
	
	/*
	 * the only way to be sure is to do it ourselves
	 */
	hostname = "unknown";	/* assume the worst */
	hostbuf[0] = 0;
	i = sizeof(from);
	if (getpeername(s, (struct sockaddr *)&from, &i) >= 0) {
#if 1
		caller = ipCallerId(&from, 1|8, &faked); /* paranoid lookup */
		if (faked == 0) {
			hostname = strncpy(hostbuf, caller,
					   sizeof(hostbuf) - 1);
			if (cp = strchr(hostbuf, '/')) {
				*cp++ = '\0';
			}
		} else {
			(void) strncpy(hostbuf, faked, sizeof(hostbuf) - 1);
			if (cp = strchr(caller, '/')) {
				++cp;
			}
		}
		if (cp) {
			(void) strncpy(hostaddr, cp, sizeof(hostaddr) - 1);
		} else {
			/* very suss! */
			(void) strncpy(hostaddr, inet_ntoa(from.sin_addr),
				       sizeof(hostaddr) - 1);
		}
		hostaddr[sizeof(hostaddr) - 1] = 0;	/* just in case */
#else
		(void) strncpy(hostaddr, inet_ntoa(from.sin_addr),
			       sizeof(hostaddr) - 1);
		hostaddr[sizeof(hostaddr) - 1] = 0;	/* just in case */

		/*
		 * Since we never issue certificates to "localhost", it will
		 * never succeed.  Try and resolve it to a name.
		 */
		if (strcmp(hostaddr, "127.0.0.1") == 0) {
			hostname = strncpy(hostbuf, (char *) Hostname(1),
					   sizeof(hostbuf) - 1);
		} else {
			if (hp = gethostbyaddr((char *)&from.sin_addr,
					       sizeof (struct in_addr),
					       from.sin_family)) {
				(void) strncpy(hostbuf, hp->h_name,
					       sizeof(hostbuf) - 2);
			}
			/*
			 * For a host cert, we do care about the name,
			 * leave it at "unknown" though, unless
			 * gethostby{addr,name} agree not only on the
			 * address list, but the fully qualified name.
			 */
			if ((hp = gethostbyname(hostbuf)) != 0
			    && strcmp(hostbuf, hp->h_name) == 0) {
			
				for (; ; hp->h_addr_list++) {
					if (hp->h_addr_list[0] == NULL) {
						syslog(LOG_NOTICE|LOG_AUTH,
						       "Address %s, not listed for host %s",
						       hostaddr,
						       hp->h_name);
						break;
					}
					if (memcmp(hp->h_addr_list[0],
						   (caddr_t)&from.sin_addr,
						   sizeof(from.sin_addr)) == 0) {
						
						hostname = hostbuf;
						break;
					}
				}
			} else if (hp != 0) {
				/*
				 * Tell them - incase they are not using
				 * tcpd or libwrap.
				 */
				syslog(LOG_NOTICE|LOG_AUTH,
				       "warning: hostbyaddr(%s) != hostbyname(%s) = %s",
				       hostaddr, hostbuf, hp->h_name);
			}
			if (hostname != hostbuf)
				faked = hostbuf;
		}
#endif
	} else {
		syslog(LOG_DEBUG, "cannot getpeername(%d): %m", s);
		hostaddr[0] = 0;
	}
	hostbuf[sizeof(hostbuf) - 1] = 0;	/* just in case */

	if (cert_revoked) {
		ssl_verify_err = "cert REVOKED";
		ssl_auth_cert = ssl_verify_errcert;
		cert_name = get_x509_name((char *)0, 0);
		ssl_auth_cert = 0;
		sts = SSLD_AUTH_REVOKED;
	} else if (cert_name = get_x509_name((char *)0, 0)) {
		ssl_verify_err = 0;
		/*
		 * we are using the convention that the CN
		 * field contains either a fully qualified
		 * hostname for a server cert or a user-name.
		 * If x509_name contains a '.', check if its a
		 * hostname, and if so, it better match the
		 * host using it.
		 */
		if (strchr(cert_name, '.')) { /* contains a . */
			/*
			 * domaincmp(domain,spec,0):
			 * If spec is one of domain, .domain, *domain
			 * or *.domain this will match.
			 * .domain, *domain or *.domain could
			 * result in accepting certs from hosts we did
			 * not intend, so by default, we use strcasecmp().
			 */
			if (domaincmp(hostname, cert_name, 0)) {
				/*
				 * does not match, was it really a
				 * host cert?
				 */
				if (strstr(cert_name, "*.") ||
				    gethostbyname(cert_name)) {
					syslog(LOG_NOTICE|LOG_AUTH,
					       "/CN=%s not %s/%s, denied",
					       cert_name, hostname, hostaddr);
					ssl_verify_err = "wrong hostcert";
					sts = SSLD_AUTH_WRONG_HOST;
				}
			}
		}
		if (ssl_verify_err == 0) {
			/*
			 * Ok, all we really care about is
			 * whether the given cert is allowed
			 * access to key.
			 */
			if ((sts = ssld_auth_cert(key, ssl_auth_cert, hostname, 0)) < 0) {
				switch(sts) {
				case SSLD_AUTH_WRONG_HOST:
					sprintf(msg, "cert denied permission for '%s' from %s/%s",
					key, hostname, hostaddr);
					break;
				default:
					sprintf(msg, "cert denied permission for '%s'",
						key);
					break;
				}
				ssl_verify_err = strdup(msg);
			}
		}
		ssl_verify_errcert = ssl_auth_cert;
	} else {
		cert_name = "";
		(void)strcpy(msg, "no cert supplied");
	}
	if (sts < 0 && ssl_verify_err != 0) {
		(void) sprintf(msg,
			       "\r\n%s: ssld_auth: %s: %.*s\r\n",
			       Hostname(1),
			       ssl_verify_err,
			       sizeof(msg)/2, ssl_verify_errcert);
		ssld_reply(msg);
		sprintf(msg, "DENIED (%.*s)", sizeof(msg) - 10, ssl_verify_err);
	} else if (sts == 0)
		sprintf(msg, "no %s file", _path_ssl_users);

	if (faked) {
		syslog(LOG_INFO|LOG_AUTH, "%s@%s (%s?) (/CN=%s) as %s: %s",
		       remuser, hostaddr, faked, cert_name, key,
		       (sts > 0) ? "OK" : msg);
	} else {
		syslog(LOG_INFO|LOG_AUTH, "%s@%s/%s (/CN=%s) as %s: %s",
		       remuser, hostname, hostaddr, cert_name, key,
		       (sts > 0) ? "OK" : msg);
	}
	return (sts);
}

static int
#ifdef SSLEAY8
ssl_verify_cb(ok, ctx)
	int ok;
	X509_STORE_CTX *ctx;
{
	const char *xs;
	int depth, error;
	static int error0 = 0;
	char *name;

	xs = (char *) X509_STORE_CTX_get_current_cert(ctx);
	error = ctx->error;
	depth = ctx->error_depth;
#else
ssl_verify_cb(ok, xs, xi, depth, error)
	int ok;
	const char *xs, *xi;
	int depth, error;
{
	static int error0 = 0;
	SSL_CTX *ssl_ctx, *sslfd_ctx();
	CERTIFICATE_CTX *ctx = 0;
	char *name;

	if ((ssl_ctx = sslfd_ctx()) != 0)
		ctx = ssl_ctx->cert;
#endif
	
	name = X509_NAME_ONELINE(X509_get_subject_name((X509 *)xs));

	/*
	 * If the verification fails, then don't remember the name.  However,
	 * if we don't require a certificate, then return success which will
	 * still allow us to set up an encrypted session. So at least we
	 * can tell peer what's wrong.
	 *
	 */
	if (!ok) {
		/* If we can't verify the issuer, then don't accept the name. */
		error0 = error;			/* say why */
		if (depth != 0 && ssl_auth_cert) {
			free(ssl_auth_cert);
			ssl_auth_cert = 0;
		}
		ssl_verify_err = X509_cert_verify_error_string(error);
		syslog(LOG_NOTICE, "verify: ok=%d,depth=%d,err=%s,xs=%s",
		       ok, depth, ssl_verify_err, name);
		if (!ssl_verify_errcert)
			ssl_verify_errcert = name;
		else
			free(name);
		
		return (ssl_verify_flag <= SSL_VERIFY_PEER);
	} else {
		char *issuer = X509_NAME_ONELINE(X509_get_issuer_name((X509 *)xs));

#ifndef SSLEAY8
		if (ssl_verify_depth > 0 && depth > ssl_verify_depth) {
			syslog(LOG_NOTICE, "verify: too deep: depth=%d,xs=%s",
			       depth, name);
			ssl_verify_err = "verify depth exceeded";
			goto verify_failed;	/* yes a goto, its ok */
		}
#endif
		switch(check_crl(ctx, xs)) {
		case 0:				/* no CRL, guess they don't care */
			syslog(LOG_DEBUG, "verify: no CRL: depth=%d,xi=%s",
			       depth, issuer);
			break;
		case -1:			/* expired CRL */
			/* but not revoked */
			syslog(LOG_NOTICE, "verify: expired CRL: depth=%d,xi=%s",
			       depth, issuer);
			break;
		case 1:				/* OK */
			break;
		default:			/* revoked */
			cert_revoked++;
			ssl_verify_err = "cert REVOKED";
			syslog(LOG_NOTICE, "verify: cert REVOKED: depth=%d,xi=%s,xs=%s",
			       depth, issuer, name);
		verify_failed:
			if (!ssl_verify_errcert)
				ssl_verify_errcert = name;
			else {
				free(name);
				name = 0;
			}
			ssl_auth_cert = 0;	/* to be sure */
			free(issuer);
			return (ssl_verify_flag <= SSL_VERIFY_PEER);
		}
		free(issuer);
	}
	syslog(LOG_INFO, "verify: depth=%d,xs=%s", depth, name);
	if (depth == 0) {
		/*
		 * In general we allow the SSL session to be
		 * established unless SSL_VERIFY_FAIL_IF_NO_PEER_CERT
		 * is set.
		 * However, we do not always accept a cert for
		 * authentication...
		 * This allows us to return an error to user.
		 */
		switch (error0) {
		case 0:				/* no error */
			ssl_auth_cert = name;
			break;
		case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
#ifdef SSLEAY8
		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
#endif
			if (ssl_verify_flag <= SSL_VERIFY_PEER)
				ssl_auth_cert = name;
			else {
				if (name)
					free(name);
				ssl_auth_cert = 0;	/* to be sure */
			}
			break;
		case X509_V_ERR_CERT_SIGNATURE_FAILURE:
		case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
		case X509_V_ERR_CERT_NOT_YET_VALID:
		case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
		case X509_V_ERR_CERT_HAS_EXPIRED:
		default:
			if (name)
				free(name);
			ssl_auth_cert = 0;	/* to be sure */
			break;
		}
	} else if (name)
		free(name);
	return ok;
}

int
ssld_init(cert_file, key_file, cipher_list, verify_flag, flags)
	const char *cert_file, *key_file, *cipher_list;
	int verify_flag;
	int flags;
{

	return ssl_init(cert_file, key_file, cipher_list, verify_flag,
			ssl_verify_cb, LOG_NOTICE, flags);
}

int
ssld_start(s, log_lev, active, cb, vf, reply)
	int s;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
	void (*reply) ();
{
	int ovf;
	int ret;

	/*
	 * our verify_cb() looks at ssl_verify_flag
	 * so we need to frob it.
	 */
	ovf = ssl_verify_flag;
	if (vf < 0)
		vf = ssl_verify_flag;
	ssl_verify_flag = (vf < 0) ? vf : (vf & 1);
	if (!cb)
		cb = ssl_verify_cb;
	set_ssldreply(reply);
	if ((ret = ssl_start_cb(s, log_lev, active, cb,
				ssl_verify_flag)) > 0) {
		char msg[512];

		msg[0] = 0;
		ssl_verify_flag = ovf;		/* restore it */
		
		if (!ssl_auth_cert) {
			/*
			 * we actually failed...
			 */
			 if (ssl_verify_err) {
			
				 (void) sprintf(msg,
						"\r\nssld_auth: %s: %.*s\r\n",
						ssl_verify_err,
						sizeof(msg)/2, ssl_verify_errcert);
				 ssld_reply(msg);
				 syslog(LOG_INFO|LOG_NOTICE, "SSL: %s: %.*s",
					ssl_verify_err,
					sizeof(msg)/2, ssl_verify_errcert);
			 }
			 if (vf & SSL_VERIFY_FAIL_IF_NO_PEER_CERT) {
				 ssl_close(s);
				 ret = -1;
			 }
		 }
	}
	return ret;
}

#ifdef MAIN
int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int Debug, optind;
	extern char *optarg;
	extern char *sock_id();
	struct sockaddr_in me;
	int 	s;
	int 	c;
	int	active = 1;			/* we don't need cert */
	char	*hostname, *cp;
	int	local_port = -1;
	const char	*logfac = "local1";
	const char	*remuser, *locuser;

	ssl_auth_cert = 0;
	remuser = locuser = 0;
	Debug = 1;				/* stop inetd_init() from forking */
	while ((c = getopt(argc, argv, "c:ar:l:z:p:L:")) != EOF) {
		switch (c) {
		case 'a':
			active = 0;
			break;
		case 'c':
			ssl_auth_cert = strdup(optarg);
			break;
		case 'r':
			remuser = optarg;
			break;
		case 'l':
			locuser = optarg;
			break;
		case 'L':
			logfac = optarg;
			break;
		case 'p':
			local_port = atoi(optarg);
			break;
		case 'z':
			ssl_setopt(optarg);
			break;
		}
	}
	argc -= optind;
	argv += optind;

	if (!remuser)
		remuser = getenv("USER");
	if (!locuser)
		locuser = remuser;


	openlog("ssld", LOG_PID, logfacs(logfac));
	if (ssl_auth_cert != 0) {
		ssl_fd = 1;
		c = ssld_auth_cert(locuser, ssl_auth_cert, "unknown", 0);
		exit(c != 1);
	}
	ssld_init(0, 0, 0, -1, 0);

	if ((s = inetd_init(local_port, SOCK_STREAM, 6, &me)) < 0) {
		syslog(LOG_ERR, "inetd_init(%d) failed: %m");
		exit(1);
	} else if (s > 0) {
		for (c = 0; c < 3; c++) {
			if (c != s) {
				close(c);
				dup(s);
			}
		}
		if (s >= 3)
			close(s);
		s = 0;
	}
	ssl_start(s, LOG_NOTICE, active);
	
	hostname = sock_id(s, 3);
	if (cp = strrchr(hostname, '.'))
		*cp = '\0';

	c = (ssld_auth(s, remuser, locuser) <= 0);
	exit(c);
}
#endif
